package org.apache.ojb.broker.util.sequence;

/* Copyright 2002-2004 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.apache.ojb.broker.PersistenceBroker;
import org.apache.ojb.broker.metadata.FieldDescriptor;

import java.util.HashMap;

/**
 * <p>
 * Very fast in memory sequence manager implementation, only the first
 * time an id was requested for a class, the manager query the database
 * for the max id in requested column - all following request were
 * performed in memory.
 * </p>
 * <b>Limitations:</b>
 * <ul type="disc">
 *	<li>do not use in clustered environments</li>
 *	<li>do not use if other applications generate id's for objects</li>
 * </ul>
 *
 *
 *
 * <p>
 * Implementation configuration properties:
 * </p>
 *
 * <table cellspacing="2" cellpadding="2" border="3" frame="box">
 * <tr>
 *     <td><strong>Property Key</strong></td>
 *     <td><strong>Property Values</strong></td>
 * </tr>
 * <tr>
 *     <td>autoNaming</td>
 *     <td>
 *          Default was 'true'. If set 'true' OJB try to build a
 *          sequence name automatic if none found in field-descriptor
 *          and set this generated name as <code>sequence-name</code>
 *          in field-descriptor. If set 'false' OJB throws an exception
 *          if none sequence name was found in field-descriptor.
 *    </td>
 * </tr>
 * </table>
 *
 * <br/>
 * <br/>
 * <br/>
 *
 * @see org.apache.ojb.broker.util.sequence.SequenceManager
 * @see org.apache.ojb.broker.util.sequence.SequenceManagerFactory
 * @see org.apache.ojb.broker.util.sequence.SequenceManagerHelper
 *
 *
 * @author <a href="mailto:armin@codeAuLait.de">Armin Waibel</a>
 * @version $Id: SequenceManagerInMemoryImpl.java,v 1.14 2004/04/04 23:53:38 brianm Exp $
 */
public class SequenceManagerInMemoryImpl extends AbstractSequenceManager
{
    private static HashMap sequenceNameKeyMap = new HashMap();

    public SequenceManagerInMemoryImpl(PersistenceBroker broker)
    {
        super(broker);
    }

    protected long getUniqueLong(FieldDescriptor field) throws SequenceManagerException
    {
        String seqName = calculateSequenceName(field);
        // we have to be threadsafe
        synchronized (sequenceNameKeyMap)
        {
            // get id for given seq name
            Long currentId = (Long) sequenceNameKeyMap.get(seqName);
            // check - first time we search for sequence name
            if (currentId == null)
            {
                currentId = new Long(
                        SequenceManagerHelper.getMaxForExtent(getBrokerForClass(), field));
            }
            currentId = new Long(currentId.longValue() + 1);
            // put new id back to map
            sequenceNameKeyMap.put(seqName, currentId);
            return currentId.intValue();
        }
    }
}
